#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#if (MSC_VER < 1300)

typedef unsigned char     uint8_t;
typedef   signed char      int8_t;
typedef unsigned short   uint16_t;
typedef   signed short    int16_t;
typedef unsigned int     uint32_t;
typedef   signed int      int32_t;
typedef unsigned __int64 uint64_t;
typedef   signed __int64  int64_t;
#else

#include <inttypes.h>

#endif

typedef struct {
	int32_t     sync;                           /*  0- 7 :  8 bits */
	int32_t     transport_error_indicator;      /*  8- 8 :  1 bit  */
	int32_t     payload_unit_start_indicator;   /*  9- 9 :  1 bit  */
	int32_t     transport_priority;             /* 10-10 :  1 bits */
	int32_t     pid;                            /* 11-23 : 13 bits */
	int32_t     transport_scrambling_control;   /* 24-25 :  2 bits */
	int32_t     adaptation_field_control;       /* 26-27 :  2 bits */
	int32_t     continuity_counter;             /* 28-31 :  4 bits */
} TS_HEADER;

typedef struct {
	TS_HEADER   hdr;
	uint8_t    *raw;
} TS_PACKET;

typedef struct {
	
	int         fd;
	
	int64_t     total;
	int64_t     offset;
	
	uint8_t    *buff;
	uint8_t    *curr;
	uint8_t    *tail;
	
	int         size;
	
	int         unit;

} TS_LOADER;

#define CHECK_PACKET_COUNT 8

static void show_usage();
static void tsmerge(char *src_a, char *src_b, char *dst);

static int setup_ts_loader(TS_LOADER *loader, char *path);
static int fill_ts_buffer(TS_LOADER *loader);
static int load_ts_packet(TS_PACKET *dst, TS_LOADER *loader);
static void teardown_ts_loader(TS_LOADER *loader);

static int sync_streams(TS_LOADER *a, TS_LOADER *b);
static int check_sync(TS_LOADER *a, TS_LOADER *b, int ax, int bx);

static int recover_bit_error(int dfd, TS_PACKET *p0, TS_PACKET *p1, TS_LOADER *tl0, TS_LOADER *tl1);
static int recover_drop(int dfd, TS_PACKET *p0, TS_PACKET *p1, TS_LOADER *tl0, TS_LOADER *tl1);
static int check_null_disturbance(TS_PACKET *p0, TS_PACKET *p1, TS_LOADER *tl0, TS_LOADER *tl1);

static void show_progress_status(int64_t offset, int64_t total);
static void show_finish();

static int select_unit_size(unsigned char *head, unsigned char *tail);

static int load_ts_packet(TS_PACKET *dst, TS_LOADER *src);
static void extract_ts_header(TS_HEADER *dst, uint8_t *src);
static unsigned char *resync(unsigned char *head, unsigned char *tail, int unit_size);
static int check_unit_invert(unsigned char *head, unsigned char *tail);
static int count_ff(uint8_t *head, uint8_t *tail);
static int skip_tot(TS_LOADER *a, TS_LOADER *b,int dfd);
static void Convert_mjd_to_date(unsigned short mjd,long *p_year,long *p_month,long *p_day);

int main(int argc, char **argv)
{
	if(argc < 4){
		show_usage();
		exit(EXIT_FAILURE);
	}

	tsmerge(argv[1], argv[2], argv[3]);
}

static void show_usage()
{
	fprintf(stderr, "tsmerge - Non-error TS packets selecter ver.0.1.1r2\n");
	fprintf(stderr, "usage: tsmerge src_a.ts src_b.ts dst.ts\n");
}

static void tsmerge(char *src_a, char *src_b, char *dst)
{
	TS_LOADER tl0;
	TS_LOADER tl1;

	TS_PACKET p0,p1;

	int n;
	int dfd;

	int64_t old_offset;
	
	memset(&tl0, 0, sizeof(TS_LOADER));
	memset(&tl1, 0, sizeof(TS_LOADER));
	
	tl0.fd = -1;
	tl1.fd = -1;
	dfd = -1;

	n = setup_ts_loader(&tl0, src_a);
	if(n < 0){
		fprintf(stderr, "error - failed on setup_ts_loader(%s) : src_a : code=%d\n", src_a, n);
		goto LAST;
	}

	n = setup_ts_loader(&tl1, src_b);
	if(n < 0){
		fprintf(stderr, "error - failed on setup_ts_loader(%s) : src_b : code=%d\n", src_b, n);
		goto LAST;
	}

	dfd = _open(dst, _O_WRONLY|_O_BINARY|_O_CREAT|_O_TRUNC, _S_IREAD|_S_IWRITE);
	if(dfd < 0){
		fprintf(stderr, "error - failed on open(%s) : dst\n", dst);
		goto LAST;
	}

	n = sync_streams(&tl0, &tl1);
	if(n < 0){
		fprintf(stderr, "error - failed on sync_streams() : code=%d\n", n);
		teardown_ts_loader(&tl0);
		teardown_ts_loader(&tl1);
		setup_ts_loader(&tl0, src_a);
		setup_ts_loader(&tl1, src_b);
		if(skip_tot(&tl0,&tl1,dfd) < 0) {
			fprintf(stderr, "error - failed on skip_tot()\n");
			goto LAST;
		}
		//goto LAST;
	}

	old_offset = 0;

	while( (load_ts_packet(&p0, &tl0) >= 0) &&
	       (load_ts_packet(&p1, &tl1) >= 0) ){
		if( (p0.hdr.transport_error_indicator != 0) ||
		    (p1.hdr.transport_error_indicator != 0) ){
			/* bit error */
			n = recover_bit_error(dfd, &p0, &p1, &tl0, &tl1);
			if(n < 0){
				fprintf(stderr, "error - failed on recover_bit_error() : code=%d\n", n);
				if(skip_tot(&tl0,&tl1,dfd) < 0) {
					fprintf(stderr, "error - failed on skip_tot()\n");
					goto LAST;
				}
				//goto LAST;
			}
		}else if(memcmp(p0.raw, p1.raw, 188) != 0){
			/* drop */
			n = recover_drop(dfd, &p0, &p1, &tl0, &tl1);
			if(n < 0){
				fprintf(stderr, "error - failed on recover_drop() : code=%d\n", n);
				if(skip_tot(&tl0,&tl1,dfd) < 0) {
					fprintf(stderr, "error - failed on skip_tot()\n");
					goto LAST;
				}
				//goto LAST;
			}
		}else{
			/* no error */
			n = _write(dfd, p0.raw, 188);
			if(n <= 0){
				fprintf(stderr, "error - failed on _write()\n");
				goto LAST;
			}
		}
		if(old_offset != tl0.offset){
			show_progress_status(tl0.offset, tl0.total);
			old_offset = tl0.offset;
		}
	}

	show_finish();
	
LAST:
	teardown_ts_loader(&tl0);
	teardown_ts_loader(&tl1);
	
	if(dfd >= 0){
		_close(dfd);
		dfd = -1;
	}
}

static void Convert_mjd_to_date(unsigned short mjd,long *p_year,long *p_month,long *p_day)
{
	double ajd = (double)mjd + 2400000.5;
	double t=ajd+0.5;
	long   z=(long)t;
	double f=t-z;
	long a;
	long x;
	double b,c,d,e;
	if(z<2299161)
	{
		a=z;
	} else {
		x=(long)((z-1867216.25)/36524.25);
		a=z+1+x-(x/4); 
	}
	b=a+1524;
	c=(long)((b-122.1)/365.25);
	d=(long)(365.25*c);
	e=(long)((b-d)/30.6001);
	*p_day  =(long)(b-d-(long)(30.6001*e)+f);
	*p_month=(long)(e<14?e-1:e-13);
	*p_year =(long)(*p_month>2?c-4716:c-4715);
}

static int skip_tot(TS_LOADER *a, TS_LOADER *b,int dfd)
{
	int n;
	TS_PACKET p0,p1;
	time_t tot0 = 0,tot1 = 0;
	long year,month,day;
	struct tm etim0;
	struct tm etim1;

	while(1) {
		if(tot0 == 0) {
			if(load_ts_packet(&p0, a) >= 0) {
				n = _write(dfd, p0.raw, 188);
				if(n <= 0){
					fprintf(stderr, "error - failed on _write()\n");
					return -1;
				}
				if(p0.hdr.pid == 0x14 && p0.raw[5] == 0x73) {
					Convert_mjd_to_date((p0.raw[8]<<8)+p0.raw[9],&year,&month,&day);
					etim0.tm_year = year-1900;
					etim0.tm_mon =  month - 1;
					etim0.tm_mday = day;
					etim0.tm_hour = (p0.raw[10]>>4)*10+(p0.raw[10]&0x0f);
					etim0.tm_min = (p0.raw[11]>>4)*10+(p0.raw[11]&0x0f);
					etim0.tm_sec = (p0.raw[12]>>4)*10+(p0.raw[12]&0x0f);
					tot0 = mktime(&etim0);
				}
			}
			else
				return -2;
		}
		if(tot1 == 0) {
			if(load_ts_packet(&p1, b) >= 0) {
				if(p1.hdr.pid == 0x14 && p1.raw[5] == 0x73) {
					Convert_mjd_to_date((p1.raw[8]<<8)+p1.raw[9],&year,&month,&day);
					etim1.tm_year = year-1900;
					etim1.tm_mon =  month - 1;
					etim1.tm_mday = day;
					etim1.tm_hour = (p1.raw[10]>>4)*10+(p1.raw[10]&0x0f);
					etim1.tm_min = (p1.raw[11]>>4)*10+(p1.raw[11]&0x0f);
					etim1.tm_sec = (p1.raw[12]>>4)*10+(p1.raw[12]&0x0f);
					tot1 = mktime(&etim1);
				}
			}
		}
		if(tot0 != 0 && tot1 != 0) {
			fprintf(stderr,"SkipTOT:[0]%04d/%02d/%02d %02d:%02d:%02d , [1]%04d/%02d/%02d %02d:%02d:%02d\n",etim0.tm_year+1900,etim0.tm_mon+1,etim0.tm_mday,etim0.tm_hour,etim0.tm_min,etim0.tm_sec
				,etim1.tm_year+1900,etim1.tm_mon+1,etim1.tm_mday,etim1.tm_hour,etim1.tm_min,etim1.tm_sec);
			if(tot0 == tot1)
				break;
			if(tot0 > tot1)
				tot1 = 0;
			else
				tot0 = 0;
		}
	}
	return 0;
}

// 1KB̃f[^obt@쐬t@Cf[^ǂݍ
static int setup_ts_loader(TS_LOADER *loader, char *path)
{
	int n;
	
	loader->fd = _open(path, _O_BINARY|_O_RDONLY|_O_SEQUENTIAL);
	if(loader->fd < 0){
		return -1;
	}
	_lseeki64(loader->fd, 0, SEEK_END);
	loader->total = _telli64(loader->fd);
	_lseeki64(loader->fd, 0, SEEK_SET);
	loader->offset = 0;


	n = 1024*1024;

	loader->buff = (uint8_t *)calloc(1, n);
	if(loader->buff == NULL){
		return -2;
	}

	loader->curr = loader->buff;
	loader->tail = loader->buff;
	loader->size = n;

	n = _read(loader->fd, loader->buff, loader->size);
	if(n < 1){
		return -3;
	}

	loader->tail += n;

	loader->unit = select_unit_size(loader->curr, loader->tail);
	if(loader->unit < 188){
		return -4;
	}

	return 0;
}

// obt@|C^̐擪ǂݎς݃f[^폜A݈ʒu擪ֈړ
// LOADER̋󂫃GAɃt@Cf[^ǋLǂݍ݂sB
// buff                       curr                               tail
//  |-------------------------|+++++++++++++++++++++++++++++++++++|
//                        currbuffֈړ
// buff/curr                         tail
//  |+++++++++++++++++++++++++++++++++|---------------------------|
//                         t@Cf[^ǋL
// buff/curr                                                     tail
//  |+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++|
// ߂l͐Vǉf[^TCY
static int fill_ts_buffer(TS_LOADER *loader)
{
	int m,n;

	m = loader->curr - loader->buff;
	n = loader->tail - loader->curr;

	loader->offset += m;
	memcpy(loader->buff, loader->curr, n);
	loader->curr = loader->buff;
	loader->tail = loader->buff + n;
	m = loader->size - n;
	m = _read(loader->fd, loader->tail, m);
	loader->tail += m;

	return m;
}

// LOADER1pPbg擾
static int load_ts_packet(TS_PACKET *dst, TS_LOADER *loader)
{
	int m,n;

	// c肪200oCgcurr擪ֈړAɃt@Cf[^ǋLB
	n = loader->tail - loader->curr;
	if(n < 200){
		m = fill_ts_buffer(loader);
		n = loader->tail - loader->curr;
	}

	// ̂f[^188oCgȉȂG[
	if(n < 188){
		return -1;
	}
	//Ō1pPbg
	if(n == 188){
		if(loader->curr[0] != 0x47){
			return -2;
		}
		extract_ts_header(&(dst->hdr), loader->curr);
		dst->raw = loader->curr;
		loader->curr += loader->unit;
		return 0;
	}
	// 擪́ApPbg擪47hŖꍇresyncs
	if( (loader->curr[0] != 0x47) ||
	    (loader->curr[loader->unit] != 0x47) ){
		uint8_t *p;
		p = resync(loader->curr, loader->tail, loader->unit);
		if(p == NULL){
			m = fill_ts_buffer(loader);
			if(m < 1){
				return -3;
			}
			p = resync(loader->curr, loader->tail, loader->unit);
			if(p == NULL){
				return -4;
			}
		}
		loader->curr = p;
	}

	extract_ts_header(&(dst->hdr), loader->curr);
	dst->raw = loader->curr;
	loader->curr += loader->unit;

	return 0;
}
// LOADEȐI
static void teardown_ts_loader(TS_LOADER *loader)
{
	if(loader->fd >= 0){
		_close(loader->fd);
		loader->fd = -1;
	}

	if(loader->buff != NULL){
		free(loader->buff);
		loader->buff = NULL;
	}

	loader->curr = NULL;
	loader->tail = NULL;

	loader->size = 0;
	loader->unit = 0;
}

static int sync_streams(TS_LOADER *a, TS_LOADER *b)
{
	int i,j,m,n;
	int ok,ng;
	int sz0,sz1;
	uint8_t *p0,*p1;

	int64_t start;
	int64_t limit;
	int64_t offset;
	int64_t pos0,pos1,diff;

	TS_HEADER hdr;

	sz0 = a->unit * CHECK_PACKET_COUNT;
	sz1 = b->unit * CHECK_PACKET_COUNT;

	if( (a == NULL) || (b == NULL) ){
		return -1;
	}
	if( (a->fd < 0) || (b->fd < 0) ){
		return -2;
	}
	if( (a->buff == NULL) || (b->buff == NULL) ){
		return -3;
	}
	if( (sz0 > a->size) || (sz1 > b->size) ){
		return -4;
	}

	start = (a->total >> 6);
	limit = start * 2;

	_lseeki64(a->fd, start, SEEK_SET);
	n = _read(a->fd, a->buff, a->size);
	a->curr = a->buff;
	a->tail = a->buff+n;
	
	p0 = resync(a->curr, a->tail, a->unit);
	while( (p0 + sz0) <= a->tail ){
		for(i=0;i<CHECK_PACKET_COUNT;i++){
			extract_ts_header(&hdr, p0 + i*(a->unit));
			if(hdr.transport_error_indicator != 0){
				break;
			}
		}
		if(i == CHECK_PACKET_COUNT){
			break;
		}
		p0 += (i+1)*(a->unit);
		if( (p0[0] != 0x47) || (p0[a->unit] != 0x47) ){
			uint8_t *p;
			p = resync(p0, a->tail, a->unit);
			if(p == NULL){
				return -5;
			}
			p0 = p;
		}
	}

	if( (p0+sz0) > a->tail ){
		return -6;
	}

	_lseeki64(b->fd, 0, SEEK_SET);
	n = _read(b->fd, b->buff, b->size);
	b->curr = b->buff;
	b->tail = b->buff + n;
	offset = 0;
	p1 = resync(b->curr, b->tail, b->unit);
	while( offset < limit ){
		while( (p1 + sz1) <= b->tail ){
			
			ok = 0;
			ng = 0;
			for(i=0;i<CHECK_PACKET_COUNT;i++){
				extract_ts_header(&hdr, p1 + i*(b->unit));
				if(hdr.transport_error_indicator != 0){
					ng += 1;
					continue;
				}
				if(memcmp(p0 + i*(a->unit), p1 + i*(b->unit), 188) == 0){
					ok += 1;
					continue;
				}
				break;
			}
			
			if( (ok > 3) && ((ng+ok) == CHECK_PACKET_COUNT) ){
				goto PHASE_2;
			}

			p1 += b->unit;
			if( (p1[0] != 0x47) || (p1[b->unit] != 0x47) ){
				uint8_t *p;
				p = resync(p1, b->tail, b->unit);
				if(p == NULL){
					break;
				}
				p1 = p;
			}
		}

		n = b->tail - p1;
		memcpy(b->buff, p1, n);
		m = b->size - n;
		m = _read(b->fd, b->buff+n, m);

		b->curr = b->buff;
		b->tail = b->buff + (m+n);
		p1 = b->buff;

		offset += n;
		if(m == 0){
			break;
		}
	}

	return -7;

PHASE_2:

	pos0 = _telli64(a->fd) - (a->tail - p0);
	pos1 = _telli64(b->fd) - (b->tail - p1);
	
	diff = pos0 - pos1;

	if(diff < 0){
		pos0 = 0;
		pos1 = (0-diff);
	}else{
		pos0 = diff;
		pos1 = 0;
	}

	_lseeki64(a->fd, pos0, SEEK_SET);
	a->offset = pos0;
	m = _read(a->fd, a->buff, a->size);
	a->curr = a->buff;
	a->tail = a->buff + m;
	
	_lseeki64(b->fd, pos1, SEEK_SET);
	b->offset = pos1;
	m = _read(b->fd, b->buff, b->size);
	b->curr = b->buff;
	b->tail = b->buff + m;

	for(i=0;i<32;i++){
		for(j=0;j<32;j++){
			n = check_sync(a, b, i, j);
			if(n >= 0){
				a->curr += (i*a->unit);
				b->curr += (j*b->unit);
				return 0;
			}
		}
	}

	return -7;
}

int check_sync(TS_LOADER *a, TS_LOADER *b, int ax, int bx)
{
	int i;
	int ok,ng;

	uint8_t *p0;
	uint8_t *p1;

	TS_HEADER hdr0,hdr1;
	
	p0 = a->curr + (ax * a->unit);
	p1 = b->curr + (bx * b->unit);

	ng = 0;
	ok = 0;

	for(i=0;i<8;i++){

		if( p0+a->unit >= a->tail ){
			return -1;
		}

		if( p1+b->unit >= b->tail ){
			return -2;
		}
		
		if( (p0[0] != 0x47) || (p0[a->unit] != 0x47) ){
			p0 = resync(p0, a->tail, a->unit);
			if(p0 == NULL){
				return -3;
			}
		}

		if( (p1[0] != 0x47) || (p1[b->unit] != 0x47) ){
			p1 = resync(p1, b->tail, b->unit);
			if(p1 == NULL){
				return -4;
			}
		}

		extract_ts_header(&hdr0, p0);
		extract_ts_header(&hdr1, p1);

		if( (hdr0.transport_error_indicator != 0) ||
		    (hdr1.transport_error_indicator != 0) ){
			ng += 1;
		}else if(memcmp(p0, p1, 188) != 0){
			return -5;
		}else{
			ok += 1;
		}

		p0 += a->unit;
		p1 += b->unit;
		
	}

	if(ok < 6){
		return -6;
	}

	return 0;
}

static int recover_bit_error(int dfd, TS_PACKET *p0, TS_PACKET *p1, TS_LOADER *tl0, TS_LOADER *tl1)
{
	int i;
	int diff;

	int64_t ofst0,ofst1;

	static const char *ok_ng[] = {
		"oo oo oo oo",
		"xx xx xx xx",
	};

	diff = 0;
	ofst0 = _telli64(tl0->fd) - (tl0->tail - tl0->curr) - tl0->unit;
	ofst1 = _telli64(tl1->fd) - (tl1->tail - tl1->curr) - tl1->unit;

	fprintf(stdout, "----\n");
	fprintf(stdout, "detect bit error : offset[0]=%I64d, offset[1]=%I64d\n", ofst0, ofst1);
	fprintf(stdout, "  %s : %s\n", ok_ng[p0->hdr.transport_error_indicator], ok_ng[p1->hdr.transport_error_indicator]);
	for(i=0;i<188;i+=4){
		if(memcmp(p0->raw+i, p1->raw+i, 4) != 0){
			fprintf(stdout, "x ");
			diff += 1;
		}else{
			fprintf(stdout, "  ");
		}
		fprintf(stdout, "%02x %02x %02x %02x : %02x %02x %02x %02x\n",
			p0->raw[i+0], p0->raw[i+1], p0->raw[i+2], p0->raw[i+3],
			p1->raw[i+0], p1->raw[i+1], p1->raw[i+2], p1->raw[i+3]);
	}

	if( diff > 40 ){

		int side;
		side = check_null_disturbance(p0, p1, tl0, tl1);
		switch(side){
		case 0: /* p0: NULL stream pid error packet */
			tl1->curr -= tl1->unit;
			return 0;
		case 1: /* p1: NULL stream pid error packet */
			tl0->curr -= tl0->unit;
			return 0;
		}
	}

	if( p0->hdr.transport_error_indicator == 0 ){
		_write(dfd, p0->raw, 188);
		return 0;
	}

	if( p1->hdr.transport_error_indicator == 0 ){
		_write(dfd, p1->raw, 188);
		return 0;
	}

	fprintf(stdout, "  !!bit error collision!!\n");
	return -1;
}

static int recover_drop(int dfd, TS_PACKET *p0, TS_PACKET *p1, TS_LOADER *tl0, TS_LOADER *tl1)
{
	int i,m,n;
	
	int64_t ofst0,ofst1;

	uint8_t *sync;
	TS_LOADER *tl;

	tl0->curr -= tl0->unit;
	tl1->curr -= tl1->unit;

	ofst0 = _telli64(tl0->fd) - (tl0->tail - tl0->curr) - tl0->unit;
	ofst1 = _telli64(tl1->fd) - (tl1->tail - tl1->curr) - tl1->unit;

	fprintf(stdout, "----\n");
	fprintf(stdout, "detect drop : offset[0]=%I64d, offset[1]=%I64d\n", ofst0, ofst1);
	
	m = tl0->tail - tl0->curr;
	if(m < (tl0->unit * 256)){
		n = fill_ts_buffer(tl0);
		if(n == 0){
			fprintf(stdout, "  too few remain data - side-0\n");
			return -1;
		}
	}

	m = tl1->tail - tl1->curr;
	if(m < (tl1->unit * 256)){
		n = fill_ts_buffer(tl1);
		if(n == 0){
			fprintf(stdout, "  too few remain data - side-1\n");
			return -2;
		}
	}

	for(i=1;i<128;i++){
		n = check_sync(tl0, tl1, 0, i);
		if(n >= 0){
			sync = tl1->curr + (i*tl1->unit);
			tl = tl1;
			fprintf(stdout, "  %d packets are droped - side-0\n", i);
			goto PHASE_2;
		}
	}

	for(i=1;i<128;i++){
		n = check_sync(tl0, tl1, i, 0);
		if(n >= 0){
			sync = tl0->curr + (i*tl0->unit);
			tl = tl0;
			fprintf(stdout, "  %d packets are droped - side-1\n", i);
			goto PHASE_2;
		}
	}

	fprintf(stdout, "  !!sync between streams can't recover!!\n");
	return -3;

PHASE_2:
	while(tl->curr < sync){
		load_ts_packet(p0, tl);
		n = _write(dfd, p0->raw, 188);
		if(n <= 0){
			fprintf(stderr, "error - failed on _write()\n");
			return -4;
		}
	}

	return 0;
}

static int check_null_disturbance(TS_PACKET *p0, TS_PACKET *p1, TS_LOADER *tl0, TS_LOADER *tl1)
{
	int i,m,n;
	int ffc;
	int side;
	int same;

	uint8_t *p;
	
	TS_LOADER *ng;
	TS_PACKET *ok;
	
	if( (p0->hdr.transport_error_indicator != 0) &&
	    (p1->hdr.transport_error_indicator != 0) ){
		/* both broken */
		return -1;
	}

	if( p0->hdr.transport_error_indicator != 0 ){
		ffc = count_ff(p0->raw, p0->raw+188);
		ng = tl0;
		ok = p1;
		side = 0;
	}else{
		ffc = count_ff(p1->raw, p1->raw+188);
		ng = tl1;
		ok = p0;
		side = 1;
	}

	if(ffc > 160){
		return side;
	}

	m = ng->tail - ng->curr;
	if(m < (ng->unit*16)){
		ng->curr -= ng->unit;
		n = fill_ts_buffer(ng);
		ng->curr += ng->unit;
		if(n < 1){
			/* no enough data */
			return -2;
		}
	}

	p = ng->curr;
	if( (p[0] != 0x47) || (p[ng->unit] != 0x47) ){
		p = resync(p, ng->tail, ng->unit);
		if(p == NULL){
			/* sync error */
			return -3;
		}
	}

	/* check similarity */
	same = 0;
	for(i=0;i<188;i++){
		if(ok->raw[i] == p[i]){
			same += 1;
		}
	}

	if( (ffc > 94) && (same > 94) ){
		return side;
	}

	/* error packet is not NULL disturbance */
	return -4;
}

static void show_progress_status(int64_t offset, int64_t total)
{
	int64_t w;
	w = (10000 * offset) / total;
	fprintf(stderr, "\rprocessing: %02d.%02d%%", (int)(w/100), (int)(w%100));
}

static void show_finish()
{
	fprintf(stderr, "\rprocessing: finish  \n");
}

static int select_unit_size(unsigned char *head, unsigned char *tail)
{
	int i;
	int m,n,w;
	int count[320];

	unsigned char *pre;
	unsigned char *buf;

	pre = NULL;
	buf = head;
	memset(count, 0, sizeof(count));

	// 1st step, find head 0x47
	while(buf < tail){
		if(buf[0] == 0x47){
			pre = buf;
			break;
		}
		buf += 1;
	}

	if(pre == NULL){
		return 0;
	}

	// 2nd step, count up 0x47 interval
	buf = pre + 1;
	while( buf < tail ){
		if(buf[0] == 0x47){
			m = buf - pre;
			if(m < 188){
				n = check_unit_invert(head, buf);
				if( (n >= 188) && (n < 320) ){
					count[n] += 1;
					pre = buf;
				}
			}else if(m < 320){
				count[m] += 1;
				pre = buf;
			}else{
				pre = buf;
			}
		}
		buf += 1;
	}

	// 3rd step, select maximum appeared interval
	m = 0;
	n = 0;
	for(i=188;i<320;i++){
		if(m < count[i]){
			m = count[i];
			n = i;
		}
	}

	// 4th step, verify unit_size
	w = m*n;
	if( (m < 8) || (w < ((tail-head) - (w/8))) ){
		return 0;
	}

	return n;
}

static void extract_ts_header(TS_HEADER *dst, uint8_t *src)
{
	dst->sync                         =  src[0];
	dst->transport_error_indicator    = (src[1] >> 7) & 0x01;
	dst->payload_unit_start_indicator = (src[1] >> 6) & 0x01;
	dst->transport_priority           = (src[1] >> 5) & 0x01;
	dst->pid = ((src[1] & 0x1f) << 8) | src[2];
	dst->transport_scrambling_control = (src[3] >> 6) & 0x03;
	dst->adaptation_field_control     = (src[3] >> 4) & 0x03;
	dst->continuity_counter           =  src[3]       & 0x0f;
}

static unsigned char *resync(unsigned char *head, unsigned char *tail, int unit_size)
{
	int i;
	unsigned char *buf;

	buf = head;
	tail -= unit_size * 8;
	while( buf < tail ){
		if(buf[0] == 0x47){
			for(i=1;i<8;i++){
				if(buf[unit_size*i] != 0x47){
					break;
				}
			}
			if(i == 8){
				return buf;
			}
		}
		buf += 1;
	}

	return NULL;
}

static int check_unit_invert(unsigned char *head, unsigned char *tail)
{
	unsigned char *buf;

	buf = tail-188;

	while(head <= buf){
		if(buf[0] == 0x47){
			return tail-buf;
		}
		buf -= 1;
	}

	return 0;
}

static int count_ff(uint8_t *head, uint8_t *tail)
{
	int r;
	uint8_t *p;

	p = head;
	r = 0;
	while(p < tail){
		if(p[0] == 0xff){
			r += 1;
		}
		p += 1;
	}

	return r;
}

